"""
Distributions
------------------------------------

..  automodule:: variational.distributions
        :members:
        :undoc-members:

Divergences
------------------------------------

..  automodule:: variational.divergence
        :members:
        :undoc-members:

Auto-Encoding
------------------------------------

..  automodule:: variational.auto_encoder
        :members:
        :undoc-members:

Datasets
------------------------------------

..  automodule:: variational.datasets
        :members:
        :undoc-members:

Visualisation
------------------------------------

..  automodule:: variational.visualisation
        :members:
        :undoc-members:
"""

from .version import __version__

from .auto_encoder import *
from .distributions import *
from .divergence import *
from .datasets import *
from .visualisation import *
